# CRIMSON
# Copyright (c) 2006, Stephen Fisher, Susan Davidson, and Junhyong Kim, 
# University of Pennsylvania.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA.
#
# benchDefs.py - initialize console
#
# @author  Stephen Fisher
# @version $Id: startup.py,v 1.133 2009/05/06 19:34:05 fisher Exp $

##############
# NOTICE: THESE ROUTINES WILL LIKELY NEED TO BE MODIFIED TO WORK
# CORRECTLY ON A WINDOWS MACHINE. THE FILE PATH DELIMITERS WILL LIKELY
# NEED TO BE CHANGED.
##############


# load standard definitions
from startup import *

###################################################################################################

# rTrialPaup() is used to run paup algorithms from a local copy of the data
# THIS USES EXISTING NEXUS FILES, RATHER THAN DOWNLOAD THE DATA FROM SDSC
def rTrialPaup_Exist(oFileStats, oDir, iDir, nLeaves, partition, run, qType):
    # query tree and create paup input file (*.nj.in)
    file = "l" + str(nLeaves) + ".p" + str(partition).zfill(2) + ".r" + str(run).zfill(2)
    # store relevant file paths
    iFileNexus = iDir + "nexus/" + file + ".nexus"
    iFileNewick = iDir + "newick/" + file + ".newick"
    iFilePaup = oDir + "paup.in/" + file + ".nexus"
    oFileTrees = oDir + "inter/" + file + ".nexus"  # stores intermediate trees, if relevant
    oFileNexus = oDir + "paup.out/" + file + ".nexus"
    oFileNewick = oDir + "newick.out/" + file + ".newick"
    oFileLog = oDir + "log/" + file + ".log"
    # build output string for stats summary file
    out = str(partition).zfill(2) + "\t" + str(run).zfill(2)
    # create paup input file: iFileNexus is NEXUS input without paup block, iFilePaup is same file with paup block
    execute("cp " + iFileNexus + " " + iFilePaup)
    if qType == 2:
        # Neighborhood-Joining
        writeString(iFilePaup, 1, "\n\nBEGIN PAUP;\n  nj brlens=yes breakties=random showtree=no treefile=" + oFileNexus + " replace=yes;\n  quit warntsave=no;\nEND;")
    elif qType == 3:
        # Parsimony: run for 5 days, allow for 200 max trees
        writeString(iFilePaup, 1, "\n\nBEGIN PAUP;\n  set maxtrees=200 increase=no;\n  hsearch timelimit=432000;\n  savetrees file=" + oFileTrees + " format=altnex replace=yes;\n  contree all /strict=no majrule=yes treefile=" + oFileNexus + " showtree=no replace=yes;\n  quit warntsave=no;\nEND;")
    # store start time in milliseconds 
    startTime = time(1)
    # run paup
    execute("paup -n " + iFilePaup + " -l " + oFileLog)
    # compute runtime in seconds and add to head of output string
    runTime = (time(1) - startTime) / 1000
    out = str(runTime) + "\t" + out
    # get newick from NEXUS file generated by paup 
    execute("./get_tree.pl " + oFileNexus + " " + oFileNewick)
    # compare trees
    tmp = CrimsonMain.getExec("env PYTHONPATH=python/lib/python2.4/site-packages getFpFn.py -t " + iFileNewick + " -e " + oFileNewick) 
    tmp = tmp.replace("(","")
    tmp = tmp.replace(")","")
    tmp = tmp.replace("\n","")
    out += "\t" + tmp.replace(", ","\t")
    # get paup stats
    tmp = CrimsonMain.getExec("grep matrix " + oFileLog)
    out += "\t" + tmp.split()[5]
    tmp = CrimsonMain.getExec("grep CPU " + oFileLog) 
    out += "\t" + tmp.split()[3]
    if tmp.count("sec") > 0:
       	out += "\t" + tmp.split()[8]
    else:
       	out += "\t" + tmp.split()[7]
    # get parsimony specific stats
    if qType == 1:
       	tmp = CrimsonMain.getExec("grep rearrangements " + oFileLog)
       	out += "\t" + tmp.split()[6]
       	tmp = CrimsonMain.getExec("grep best " + oFileLog)
       	out += "\t" + tmp.split()[6]
    # store output to file:
    # partition run Fp Fn numChars time cpuTime
    writeString(oFileStats, 1, out + "\n")

# rTrialPaup() is used to run paup algorithms (NJ and Parsimony)
def rTrialPaup(oFileStats, oDir, nLeaves, partition, run, qType):
	# store relevant file paths
	file = "l" + str(nLeaves) + ".p" + str(partition).zfill(2) + ".r" + str(run).zfill(2)
	iQuery = "L" + str(nLeaves) + ".P" + str(partition).zfill(2) + ".R" + str(run).zfill(2)
	iFileNexus = oDir + "sample.nexus/" + file + ".nexus"
	iFileNewick = oDir + "sample.newick/" + file + ".newick"
	iFilePaup = oDir + "paup.in/" + file + ".nexus"
	oFileTrees = oDir + "inter/" + file + ".nexus"  # stores intermediate trees, if relevant
	oFileNexus = oDir + "paup.out/" + file + ".nexus"
	oFileNewick = oDir + "newick.out/" + file + ".newick"
	oFileLog = oDir + "log/" + file + ".log"
	# build output string for stats summary file
	out = str(partition).zfill(2) + "\t" + str(run).zfill(2)
	# store start time in milliseconds, including downloading of data
	dbStartTime = time(1)
	# get paup data
	tree = runNexusQuery(iQuery, iFileNexus, 0, 0, 0, 1)
	# save newick tree structure to file
	writeString(iFileNewick, 0, tree.toNewick(0,0) + ";")
	# create paup input file: iFileNexus is NEXUS input without paup block, iFilePaup is same file with paup block
	execute("cp " + iFileNexus + " " + iFilePaup)
	if qType == 0:
		# Neighborhood-Joining
		writeString(iFilePaup, 1, "\n\nBEGIN PAUP;\n  nj brlens=yes breakties=random showtree=no treefile=" + oFileNexus + " replace=yes;\n  quit warntsave=no;\nEND;")
	elif qType == 1:
		# Parsimony: run for 5 days, allow for 200 max trees
		writeString(iFilePaup, 1, "\n\nBEGIN PAUP;\n  set maxtrees=200 increase=no;\n  hsearch timelimit=432000;\n  savetrees file=" + oFileTrees + " format=altnex replace=yes;\n  contree all /strict=no majrule=yes treefile=" + oFileNexus + " showtree=no replace=yes;\n  quit warntsave=no;\nEND;")
	# store start time in milliseconds 
	startTime = time(1)
	# run paup
	execute("paup -n " + iFilePaup + " -l " + oFileLog)
	# compute runtime in seconds and add to head of output string
	runTime = (time(1) - startTime) / 1000
	out = str(runTime) + "\t" + out
	# get newick from NEXUS file generated by paup 
	execute("./get_tree.pl " + oFileNexus + " " + oFileNewick)
	# compare trees
	tmp = CrimsonMain.getExec("env PYTHONPATH=python/lib/python2.4/site-packages ./getFpFn.py -t " + iFileNewick + " -e " + oFileNewick) 
	tmp = tmp.replace("(","")
	tmp = tmp.replace(")","")
	tmp = tmp.replace("\n","")
	out += "\t" + tmp.replace(", ","\t")
	# compute full runtime in seconds and add to head of output string
	runTime = (time(1) - dbStartTime) / 1000
	out = str(runTime) + "\t" + out
	# get paup stats
	tmp = CrimsonMain.getExec("grep matrix " + oFileLog)
	out += "\t" + tmp.split()[5]
	tmp = CrimsonMain.getExec("grep CPU " + oFileLog) 
	out += "\t" + tmp.split()[3]
	if tmp.count("sec") > 0:
		out += "\t" + tmp.split()[8]
	else:
		out += "\t" + tmp.split()[7]
	# get parsimony specific stats
	if qType == 1:
		tmp = CrimsonMain.getExec("grep rearrangements " + oFileLog)
		out += "\t" + tmp.split()[6]
		tmp = CrimsonMain.getExec("grep best " + oFileLog)
		out += "\t" + tmp.split()[6]
	# store output to file:
	# partition run Fp Fn numChars time cpuTime
	writeString(oFileStats, 1, out + "\n")

# rTrial() is used to run RAxML
def rTrialRAxML_Exist(oFileStats, oDir, iDir, nLeaves, partition, run):
	# store relevant file paths
	# use leaves in label so multiple RAxML runs can share the same directory
	label = "l" + str(nLeaves) + ".p" + str(partition).zfill(2) + ".r" + str(run).zfill(2)
	iFilePhylip = iDir + "phylip/" + label + ".phylip"
	iFileNewick = iDir + "newick/" + label + ".newick"
	oFileNewick = "RAxML_bestTree." + label
	oFileLog = "RAxML_info." + label
	oFileBP = "RAxML_bipartitions." + label
	oFileBS = "RAxML_bootstrap." + label
	# build output string for stats summary file
	out = str(partition).zfill(2) + "\t" + str(run).zfill(2)
	# store start time in milliseconds 
	startTime = time(1)
	# run RAxML. The seed values will be stored in the log file
	execute("raxmlHPC -f a -x " + str(abs(random.nextInt())) + " -p " + str(abs(random.nextInt())) + " -# 100 -m GTRGAMMA -s " + iFilePhylip + " -n " + label)
	# compute runtime in seconds and add to head of output string
	runTime = (time(1) - startTime) / 1000
	out = str(runTime) + "\t" + out
	# move RAxML output files
	execute("mv " + oFileNewick + " " + oDir + "newick.out/" + oFileNewick)
	oFileNewick = oDir + "newick.out/" + oFileNewick
	execute("mv " + oFileLog + " " + oDir + "log/" + oFileLog)
	oFileLog = oDir + "log/" + oFileLog
	execute("mv " + oFileBP + " " + oDir + "bipartitions/" + oFileBP)
	execute("mv " + oFileBS + " " + oDir + "bootstrap/" + oFileBS)
	# compare trees
	tmp = CrimsonMain.getExec("env PYTHONPATH=python/lib/python2.4/site-packages ./getFpFn.py -t " + iFileNewick + " -e " + oFileNewick) 
	tmp = tmp.replace("(","")
	tmp = tmp.replace(")","")
	tmp = tmp.replace("\n","")
	out += "\t" + tmp.replace(", ","\t")
	# get number characters
	tmp = CrimsonMain.getExec("grep distinct " + oFileLog)
	out += "\t" + tmp.split()[2]
	# store output to file
	writeString(oFileStats, 1, out + "\n")

# rTrial() is used to run RAxML
def rTrialRAxML(oFileStats, oDir, nLeaves, partition, run):
	# store relevant file paths
	# use leaves in label so multiple RAxML runs can share the same directory
	label = "l" + str(nLeaves) + ".p" + str(partition).zfill(2) + ".r" + str(run).zfill(2)
	iQuery = "L" + str(nLeaves) + ".P" + str(partition).zfill(2) + ".R" + str(run).zfill(2)
	iFilePhylip = oDir + "sample.phylip/" + label + ".phylip"
	iFileNewick = oDir + "sample.newick/" + label + ".newick"
	oFileNewick = "RAxML_bestTree." + label
	oFileLog = "RAxML_info." + label
	oFileBP = "RAxML_bipartitions." + label
	oFileBS = "RAxML_bootstrap." + label
	# build output string for stats summary file
	out = str(partition).zfill(2) + "\t" + str(run).zfill(2)
	# store start time in milliseconds, including downloading of data
	dbStartTime = time(1)
	# get phylip data
	tree = runPhylipQuery(iQuery, iFilePhylip, 0)
	# save newick tree structure to file
	writeString(iFileNewick, 0, tree.toNewick(0,0) + ";")
	# store start time in milliseconds 
	startTime = time(1)
	# run RAxML. The seed values will be stored in the log file
	execute("raxmlHPC -f a -x " + str(abs(random.nextInt())) + " -p " + str(abs(random.nextInt())) + " -# 100 -m GTRGAMMA -s " + iFilePhylip + " -n " + label)
	# compute runtime in seconds and add to head of output string
	runTime = (time(1) - startTime) / 1000
	out = str(runTime) + "\t" + out
	# move RAxML output files
	execute("mv " + oFileNewick + " " + oDir + "newick.out/" + oFileNewick)
	oFileNewick = oDir + "newick.out/" + oFileNewick
	execute("mv " + oFileLog + " " + oDir + "log/" + oFileLog)
	oFileLog = oDir + "log/" + oFileLog
	execute("mv " + oFileBP + " " + oDir + "bipartitions/" + oFileBP)
	execute("mv " + oFileBS + " " + oDir + "bootstrap/" + oFileBS)
	# compare trees
	tmp = CrimsonMain.getExec("env PYTHONPATH=python/lib/python2.4/site-packages ./getFpFn.py -t " + iFileNewick + " -e " + oFileNewick) 
	tmp = tmp.replace("(","")
	tmp = tmp.replace(")","")
	tmp = tmp.replace("\n","")
	out += "\t" + tmp.replace(", ","\t")
	# compute full runtime in seconds and add to head of output string
	runTime = (time(1) - dbStartTime) / 1000
	out = str(runTime) + "\t" + out
	# get number characters
	tmp = CrimsonMain.getExec("grep distinct " + oFileLog)
	out += "\t" + tmp.split()[2]
	# store output to file
	writeString(oFileStats, 1, out + "\n")

def rSet(nRuns, nLeaves, qType, rStart, rStop, pStart, pStop):
    """rSet(nRuns, nLeaves, qType, rStart, rStop, pStart, pStop):
    If rStart and rStop are '-1' then all nRuns will be performed. If
    pStart and pStop are '-1' then all 20 partitions will be used.
    """
    rDir = "l" + str(nLeaves) + ".r" + str(nRuns)
    iDir = "data/input/" + rDir + "/"
    # Neighborhood-Joining
    if qType == 0 or qType == 2: 
       	oDir = "data/nj/" + rDir
       	execute("mkdir -p " + oDir)
       	execute("mkdir -p " + oDir + "/sample.nexus/")
       	execute("mkdir -p " + oDir + "/paup.in/")
       	execute("mkdir -p " + oDir + "/paup.out/")
    # Parsimony
    elif qType == 1 or qType == 3: 
       	oDir = "data/parsimony/" + rDir
       	execute("mkdir -p " + oDir)
       	execute("mkdir -p " + oDir + "/sample.nexus/")
       	execute("mkdir -p " + oDir + "/paup.in/")
       	execute("mkdir -p " + oDir + "/paup.out/")
       	execute("mkdir -p " + oDir + "/inter/")
    # RAxML
    elif qType == 4 or qType == 5: 
       	oDir = "data/raxml/" + rDir
       	execute("mkdir -p " + oDir)
       	execute("mkdir -p " + oDir + "/sample.phylip/")
       	execute("mkdir -p " + oDir + "/bootstrap/")
       	execute("mkdir -p " + oDir + "/bipartitions/")
    else:
       	print "ERROR: Invalid qType"
       	return
    oFileStats = oDir + ".txt"
    oDir = oDir + "/"
    execute("mkdir -p " + oDir + "sample.newick/")
    execute("mkdir -p " + oDir + "log/")
    execute("mkdir -p " + oDir + "newick.out/")
    # file containing tab delimited stats for all runs
    #if qType == 0: writeString(oFileStats, 0, "DB Time (s)\tTime (s)\tPartition\tRun\tFp\tFn\tCharacters\tPaup-time\tPaup-CPU\n")
    #if qType == 1: writeString(oFileStats, 0, "DB Time (s)\tTime (s)\tPartition\tRun\tFp\tFn\tCharacters\tPaup-time\tPaup-CPU\tRearrangements\tScore\n")
    #if qType == 2: writeString(oFileStats, 0, "DB Time (s)\tTime (s)\tPartition\tRun\tFp\tFn\tCharacters\n")
    if rStart == -1: rStart = 1
    if rStop == -1: rStop = nRuns + 1
    if pStart == -1: pStart = 1
    if pStop == -1: pStop = 21
    for run in range(rStart, rStop):
       	for partition in range(pStart, pStop):
            # process each of the 20 partitions
            time()
            print "Processing partition: " + str(partition) + "  run: " + str(run)
            if qType == 0 or qType == 1: rTrialPaup(oFileStats, oDir, nLeaves, partition, run, qType)
            elif qType == 2 or qType == 3: rTrialPaup_Exist(oFileStats, oDir, iDir, nLeaves, partition, run, qType)
            elif qType == 4: rTrialRAxML(oFileStats, oDir, nLeaves, partition, run)
            elif qType == 5: rTrialRAxML_Exist(oFileStats, oDir, iDir, nLeaves, partition, run)


###################################################################################################

def getData(oDir, nLeaves, nRuns, partition, run):
    """getData(nLeaves, nRuns, partition, run):
    This will run a specific query and download the resulting NEXUS
    and phylip files. The sampled tree's newick structures will also
    be ouptut.
    """
    # store relevant file paths
    # use leaves in label so multiple RAxML runs can share the same directory
    label = "L" + str(nLeaves) + ".p" + str(partition).zfill(2) + ".r" + str(run).zfill(2)
    iQuery = "L" + str(nLeaves) + ".P" + str(partition).zfill(2) + ".R" + str(run).zfill(2)
    iFileNexus = oDir + "nexus/" + label + ".nexus"
    iFilePhylip = oDir + "phylip/" + label + ".phylip"
    iFileNewick = oDir + "newick/" + label + ".newick"
    # get paup data
    runNexusQuery(iQuery, iFileNexus, 0, 0, 0, 0)
    # get phylip data
    tree = runPhylipQuery(iQuery, iFilePhylip, 0)
    # save newick tree structure to file
    writeString(iFileNewick, 0, tree.toNewick(0,0) + ";")

def rGetInput(oDir, nLeaves, nRuns, rStart, rEnd, pStart, pEnd):
    """rGetInput(oDir, nLeaves, nRuns, rStart, rEnd, pStart, pEnd):
    This will run a set of queries and download the resulting NEXUS,
    phylip, and newick files. The output directory 'oDir' will be
    created if it doesn't already exist. The subdirectorie
    'L<nLeaves>.R<nRuns>' and it's subdirectories 'nexus', 'phylip',
    and 'newick' will also be created, as needed.
    """
    execute("mkdir -p " + oDir)
    oDir = oDir + "/L" + str(nLeaves) + ".R" + str(nRuns).zfill(2) + "/"
    execute("mkdir -p " + oDir)
    execute("mkdir -p " + oDir + "nexus")
    execute("mkdir -p " + oDir + "phylip")
    execute("mkdir -p " + oDir + "newick")
    if rStart == -1: rStart = 1
    if rEnd == -1: rEnd = nRuns + 1
    if pStart == -1: rStart = 1
    if pEnd == -1: rEnd = 21
    for run in range(rStart, rEnd):
        for partition in range(pStart, pEnd):
            print "Processing partition: " + str(partition) + "  run: " + str(run)
            getData(oDir, nLeaves, nRuns, partition, run)


###################################################################################################

def createQsub(nLeaves, nRuns):
    """createQsub(nLeaves, nRuns):
    Used to create qsub files for running on cluster. 
    """
    for run in range(1, nRuns+1):
        for partition in range(1, 21):
            job = "jobs/parsimony-p.l"+str(nLeaves)+".r"+str(nRuns)+"/parsimony.l"+str(nLeaves) + ".r"+str(run).zfill(2) + ".p"+str(partition).zfill(2)
            txt = "rSet("+str(nRuns)+","+str(nLeaves)+",3," + str(run) + "," + str(run+1) + "," + str(partition) + "," + str(partition+1) + ")\n"
            writeString(job, 0, txt)

###################################################################################################

def siteChanges(filename):
    """siteChanges(filename):
    This expects a phylip file and will return the minimum number of
    site changes for the species included in the file.
    """
    iFile = open(filename, "r")
    if iFile == None: return
    
    # get first line
    line = iFile.readline()
    # split line at white space
    cols = line.split()
    numChars = int(cols[1])
    list = java.util.ArrayList(numChars)
    
    for i in range(0, numChars): list.add(java.util.HashSet(4))
    list = list.toArray()

    while 1:
        line = iFile.readline()
        if not line.strip(): break

        # ignore species ID
        s = line.split()[1]

        # loop through each column and add char to appropriate set
        for i in range(0, numChars): 
            if s[i] != "-": list[i].add(s[i])

    tot = 0
    for i in range(0, numChars): 
        if list[i].size() > 1: tot += list[i].size() - 1

    return tot

def computeSiteChanges(nLeaves, nRuns):
    """computeSiteChanges(nLeaves, nRuns):
    Compute the minimum number of site changes for all queries
    specified by nLeaves and nRuns. 
    """
    oFile = "data/input/l"+str(nLeaves) + ".r"+str(nRuns) + "/minSiteChanges.txt"
    writeString(oFile, 0, "partition\trun\tscore\n")
    for run in range(1, nRuns+1):
        for partition in range(1, 21):
            iFile = "data/input/l"+str(nLeaves) + ".r"+str(nRuns) + "/phylip.reduced/l"+str(nLeaves) + ".p"+str(partition).zfill(2) + ".r"+str(run).zfill(2)+".phylip.reduced"
            changes = siteChanges(iFile)
            out = str(partition).zfill(2) + "\t" + str(run).zfill(2) + "\t" + str(changes) + "\n"
            writeString(oFile, 1, out)

