# Copyright (c) 2006, Stephen Fisher, Susan Davidson, and Junhyong Kim, 
# University of Pennsylvania.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA.
#
# exBatchScript-2.py - sample batch script

# This script will run the query "L10.P03.R05" from the SDSC database,
# creating a 10 leaf sub-tree, and saving the output as a phylip file.
# The phylip file is used by RAxML to reconstruct the sub-tree.

# Load standard definitions
from startup import *

# Load benchmarking specific definitions
from benchDefs import *

# Specify that we will be connecting to an Oracle server
setDBType("oracle")

# Open a read-only connection to the SDSC Oracle server, via web proxy
openDatabase("", "", "http://www.phylo.org/DatabaseProxy.php")

# This query creates a 10 leaf tree, sampling the 1M leaf tree
# ("TREE-0"), using the data partition 3 and "6695834184220664292L" as
# the seed for the random number generator.
queryID = "L10.p03.r05"

# Using labeling for output files that is consistent with queries in
# CIPRES database: L<nLeaves>.p<partition>.r<run>.
iFilePhylip = queryID + ".phylip"

# Run query, creating a phylip file called "L10.p03.r05.phylip"
# containing the sequence data for the sampled tree. Upon completion,
# the variable "newTree" will contain the 10 leaf tree Java object.
newTree = runPhylipQuery(queryID, iFilePhylip, 0)

# Save the new tree's newick structure to the file
# "L10.p03.r05.newick," appending a ";" to the end of the newick
# string. Internal sequences will not be included and the output tree
# will be unrooted. Note that we could have called runPhylipQuery()
# with the third arg being "4" and that would have also exported the
# newick tree.
writeString(queryID + ".newick", 0, newTree.toNewick(0,0) + ";")

# Store start time for RAxML run.
startTime = time(1)

# Run RAxML. The seed values will be stored in the log file. The RAxML
# output filenames will all include the queryID.
execute("raxmlHPC -f a -x " + str(abs(random.nextInt())) + " -p " + str(abs(random.nextInt())) + " -# 100 -m GTRGAMMA -s " + iFilePhylip + " -n " + queryID)

# Print RAxML runtime in seconds. 
print "\nRAxML run time (sec): " + str(int((time(1) - startTime) / 1000))

# Parsimony score
tmp = CrimsonMain.getExec("grep Final RAxML_info." + queryID)
print "Maximum Likelihood score: " + tmp.split()[4]

# Print the minimum number of site changes in the sampled tree
print "Minimum number of site changes: " + str(siteChanges(queryID + ".phylip"))

